// Licensed to LinDB under one or more contributor
// license agreements. See the NOTICE file distributed with
// this work for additional information regarding copyright
// ownership. LinDB licenses this file to you under
// the Apache License, Version 2.0 (the "License"); you may
// not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package metadb

import (
	"errors"
	"strings"
	"sync"

	"github.com/lindb/roaring"

	"github.com/lindb/lindb/constants"
	"github.com/lindb/lindb/kv"
	"github.com/lindb/lindb/metrics"
	"github.com/lindb/lindb/pkg/strutil"
	"github.com/lindb/lindb/series/tag"
	"github.com/lindb/lindb/sql/stmt"
	"github.com/lindb/lindb/tsdb/tblstore/tagkeymeta"
)

//go:generate mockgen -source ./tag_metadata.go -destination=./tag_metadata_mock.go -package metadb

// for testing
var (
	newTagReaderFunc  = tagkeymeta.NewReader
	newTagFlusherFunc = tagkeymeta.NewFlusher
)

// TagMetadata represents the tag metadata, stores all tag values under spec tag key
type TagMetadata interface {
	// GenTagValueID generates the tag value id for spec tag key
	GenTagValueID(tagKeyID tag.KeyID, tagValue string) (uint32, error)
	// SuggestTagValues returns suggestions from given tag key id and prefix of tag value
	SuggestTagValues(tagKeyID tag.KeyID, tagValuePrefix string, limit int) []string
	// FindTagValueDsByExpr finds tag value ids by tag filter expr for spec tag key,
	// if not exist, return nil, constants.ErrNotFound, else returns tag value ids
	FindTagValueDsByExpr(tagKeyID tag.KeyID, expr stmt.TagFilter) (*roaring.Bitmap, error)
	// GetTagValueIDsForTag get tag value ids for spec tag key of metric,
	// if not exist, return nil, constants.ErrNotFound, else returns tag value ids
	GetTagValueIDsForTag(tagKeyID tag.KeyID) (tagValueIDs *roaring.Bitmap, err error)
	// CollectTagValues collects the tag values by tag value ids,
	CollectTagValues(tagKeyID tag.KeyID,
		tagValueIDs *roaring.Bitmap,
		tagValues map[uint32]string,
	) error
	// Flush flushes the memory tag metadata into kv store
	Flush() error
}

// tagMetadata implements TagMetadata interface
type tagMetadata struct {
	databaseName string
	family       kv.Family // store tag key/value data using common kv store
	mutable      *TagStore // mutable store current writeable memory store
	immutable    *TagStore // immutable need to flush into kv store

	rwMutex sync.RWMutex

	statistics *metrics.TagMetaStatistics
}

// NewTagMetadata creates a tag metadata
func NewTagMetadata(databaseName string, family kv.Family) TagMetadata {
	return &tagMetadata{
		databaseName: databaseName,
		family:       family,
		mutable:      NewTagStore(),
		statistics:   metrics.NewTagMetaStatistics(databaseName),
	}
}

// GenTagValueID generates the tag value id for spec tag key
func (m *tagMetadata) GenTagValueID(tagKeyID tag.KeyID, tagValue string) (tagValueID uint32, err error) {
	// get tag value id from memory with read lock
	m.rwMutex.RLock()
	if tagValueID0, ok := m.getTagValueIDInMem(tagKeyID, tagValue); ok {
		m.rwMutex.RUnlock()
		return tagValueID0, nil
	}
	m.rwMutex.RUnlock()

	// try load tag value id from kv store
	snapshot := m.family.GetSnapshot()
	defer snapshot.Close()

	readers, err := snapshot.FindReaders(uint32(tagKeyID))
	if err != nil {
		// find table.Reader err, return it
		m.statistics.GenTagValueIDFailures.Incr()
		return
	}
	var reader tagkeymeta.Reader
	if len(readers) > 0 {
		// found tag data in kv store, try load tag value data
		reader = newTagReaderFunc(readers)
		tagValueID, err = reader.GetTagValueID(tagKeyID, tagValue)
		if err == nil {
			// got tag value id from kv store
			return tagValueID, nil
		}
		if !errors.Is(err, constants.ErrNotFound) {
			// if load tag value id err, return it
			m.statistics.GenTagValueIDFailures.Incr()
			return
		}
	}

	// tag value not exist, need assign new tag value id for tag value with write lock
	m.rwMutex.Lock()
	defer m.rwMutex.Unlock()
	// double check, memory if exist tag value
	if tagValueID0, ok := m.getTagValueIDInMem(tagKeyID, tagValue); ok {
		return tagValueID0, nil
	}

	// assign new tag value id
	tagEntry, ok := m.mutable.Get(uint32(tagKeyID))
	if !ok {
		if reader != nil {
			// if tag data exist in kv store, need load tag value id auto sequence
			seq, err := reader.GetTagValueSeq(tagKeyID)
			if err != nil {
				m.statistics.GenTagValueIDFailures.Incr()
				return 0, err
			}
			tagEntry = newTagEntry(seq)
		} else {
			// for new tag, auto sequence start with 1
			tagEntry = newTagEntry(0)
		}
		// cache tag entry
		m.mutable.Put(uint32(tagKeyID), tagEntry)
	}

	// assign new id
	tagValueID = tagEntry.genTagValueID()
	tagEntry.addTagValue(tagValue, tagValueID)

	m.statistics.GenTagValueIDs.Incr()

	return tagValueID, nil
}

// SuggestTagValues returns suggestions from given tag key id and prefix of tag value
func (m *tagMetadata) SuggestTagValues(tagKeyID tag.KeyID, tagValuePrefix string, limit int) []string {
	result := make([]string, 0)
	m.loadTagValueIDsInMem(tagKeyID, func(tagEntry TagEntry) {
		for value := range tagEntry.getTagValues() {
			if strings.HasPrefix(value, tagValuePrefix) {
				result = append(result, value)
			}
		}
	})

	snapshot := m.family.GetSnapshot()
	defer snapshot.Close()

	readers, err := snapshot.FindReaders(uint32(tagKeyID))
	if err != nil {
		// find table.Reader err, return it
		return nil
	}
	var reader tagkeymeta.Reader
	if len(readers) > 0 {
		// found tag data in kv store, try load tag value data
		reader = newTagReaderFunc(readers)
		readerValues := reader.SuggestTagValues(tagKeyID, tagValuePrefix, limit)
		result = append(result, readerValues...)
	}
	return result
}

// FindTagValueDsByExpr finds tag value ids by tag filter expr for spec tag key,
// if not exist, return nil, constants.ErrNotFound, else returns tag value ids
func (m *tagMetadata) FindTagValueDsByExpr(tagKeyID tag.KeyID, expr stmt.TagFilter) (*roaring.Bitmap, error) {
	result := roaring.New()
	m.loadTagValueIDsInMem(tagKeyID, func(tagEntry TagEntry) {
		ids := tagEntry.findSeriesIDsByExpr(expr)
		if ids != nil {
			result.Or(ids)
		}
	})

	err := m.loadTagValueIDsInKV(tagKeyID, func(reader tagkeymeta.Reader) error {
		tagValueIDs, err := reader.FindValueIDsByExprForTagKeyID(tagKeyID, expr)
		if err != nil {
			return err
		}
		result.Or(tagValueIDs)
		return nil
	})
	if err != nil {
		return nil, err
	}
	return result, nil
}

// GetTagValueIDsForTag get tag value ids for spec tag key of metric,
// if not exist, return nil, constants.ErrNotFound, else returns tag value ids
func (m *tagMetadata) GetTagValueIDsForTag(tagKeyID tag.KeyID) (*roaring.Bitmap, error) {
	result := roaring.New()
	m.loadTagValueIDsInMem(tagKeyID, func(tagEntry TagEntry) {
		ids := tagEntry.getTagValueIDs()
		if ids != nil {
			result.Or(ids)
		}
	})

	err := m.loadTagValueIDsInKV(tagKeyID, func(reader tagkeymeta.Reader) error {
		tagValueIDs, err := reader.GetTagValueIDsForTagKeyID(tagKeyID)
		if err != nil {
			return err
		}
		result.Or(tagValueIDs)
		return nil
	})
	if err != nil {
		return nil, err
	}
	return result, nil
}

// CollectTagValues collects the tag values by tag value ids for spec tag key,
func (m *tagMetadata) CollectTagValues(tagKeyID tag.KeyID,
	tagValueIDs *roaring.Bitmap,
	tagValues map[uint32]string,
) error {
	m.loadTagValueIDsInMem(tagKeyID, func(tagEntry TagEntry) {
		if tagValueIDs.IsEmpty() {
			// if found all tag value in memory, return it(maybe load immutable memory)
			return
		}
		tagEntry.collectTagValues(tagValueIDs, tagValues)
	})

	if tagValueIDs.IsEmpty() {
		// no need collect tag value ids, returns it
		return nil
	}
	err := m.loadTagValueIDsInKV(tagKeyID, func(reader tagkeymeta.Reader) error {
		return reader.CollectTagValues(tagKeyID, tagValueIDs, tagValues)
	})
	if err != nil {
		return err
	}
	return nil
}

// Flush flushes the memory tag metadata into kv store
func (m *tagMetadata) Flush() error {
	if !m.checkFlush() {
		return nil
	}

	// flush immutable data into kv store
	fluster := m.family.NewFlusher()
	defer fluster.Release()

	tagFluster, err := newTagFlusherFunc(fluster)
	if err != nil {
		return err
	}
	if err := m.immutable.WalkEntry(func(key uint32, value TagEntry) error {
		tagValues := value.getTagValues()
		for tagValue, tagValueID := range tagValues {
			tagFluster.FlushTagValue(strutil.String2ByteSlice(tagValue), tagValueID)
		}
		if err := tagFluster.FlushTagKeyID(key, value.getTagValueIDSeq()); err != nil {
			return err
		}
		return nil
	}); err != nil {
		return err
	}
	if err := tagFluster.Close(); err != nil {
		return err
	}
	// finally, clear immutable
	m.rwMutex.Lock()
	m.immutable = nil
	m.rwMutex.Unlock()
	return nil
}

// checkFlush checks if need do flush job, if need, do switch mutable/immutable
func (m *tagMetadata) checkFlush() bool {
	m.rwMutex.Lock()
	defer m.rwMutex.Unlock()

	if m.mutable.Size() == 0 && m.immutable == nil {
		// no new data or immutable is not nil
		return false
	}
	if m.mutable.Size() > 0 && m.immutable == nil {
		// reset mutable, if flush fail immutable is not nil
		m.immutable = m.mutable
		m.mutable = NewTagStore()
	}
	return true
}

// loadTagValueIDsInKV loads tag value ids in kv store
func (m *tagMetadata) loadTagValueIDsInKV(tagKeyID tag.KeyID, fn func(reader tagkeymeta.Reader) error) error {
	// try load tag value id from kv store
	snapshot := m.family.GetSnapshot()
	defer snapshot.Close()

	readers, err := snapshot.FindReaders(uint32(tagKeyID))
	if err != nil {
		// find table.Reader err, return it
		return err
	}
	var reader tagkeymeta.Reader
	if len(readers) > 0 {
		// found tag data in kv store, try load tag value data
		reader = newTagReaderFunc(readers)
		if err := fn(reader); err != nil {
			return err
		}
	}
	return nil
}

// loadTagValueIDsInMem loads tag value ids from mutable/immutable store
func (m *tagMetadata) loadTagValueIDsInMem(tagKeyID tag.KeyID, fn func(tagEntry TagEntry)) {
	// define get tag value ids func
	getTagValueIDs := func(tagStore *TagStore) {
		if tagEntry, ok := tagStore.Get(uint32(tagKeyID)); ok {
			fn(tagEntry)
		}
	}

	m.rwMutex.RLock()
	defer m.rwMutex.RUnlock()

	getTagValueIDs(m.mutable)
	if m.immutable != nil {
		getTagValueIDs(m.immutable)
	}
}

// getTagValueIDInMem gets tag value id from mutable/immutable store
func (m *tagMetadata) getTagValueIDInMem(tagKeyID tag.KeyID, tagValue string) (tagValueID uint32, ok bool) {
	tagValueID, ok = getTagValueID(m.mutable, tagKeyID, tagValue)
	if ok {
		return
	}
	if m.immutable != nil {
		tagValueID, ok = getTagValueID(m.immutable, tagKeyID, tagValue)
		if ok {
			return
		}
	}
	return
}

// getTagValueID gets tag value id from tag store based on tag key id and tag value
func getTagValueID(tags *TagStore, tagKeyID tag.KeyID, tagValue string) (tagValueID uint32, ok bool) {
	if tagEntry, ok := tags.Get(uint32(tagKeyID)); ok {
		tagValueID, ok = tagEntry.getTagValueID(tagValue)
		if ok {
			return tagValueID, true
		}
	}
	return
}
